from __future__ import annotations
from aslabs.dependencies import Dependencies as Deps
from enum import Enum
from typing import Any, Callable, Dict, Optional, List, Sequence, Type, Union, TypeVar
from fastapi import APIRouter, params, Response, routing, FastAPI, Request, utils, Depends
from fastapi.routing import APIRoute
from starlette.routing import BaseRoute
from starlette.types import ASGIApp
from fastapi.datastructures import Default
from starlette.responses import JSONResponse
import contextvars


class DependencyRouter(APIRouter):

    _dependency_setup: Optional[Callable[[Deps], Optional[Deps]]] = None

    def register_dependencies(self, reg: Callable[[Deps], Optional[Deps]]) -> DependencyRouter:
        self._dependency_setup = reg
        return self

    def apply_dependencies(self, deps: Deps) -> Deps:
        if self._dependency_setup is None:
            return deps

        self._dependency_setup(deps)
        return deps


dependency_context = contextvars.ContextVar(
    "dependency_context", default=Deps())


class Dependencies:
    def __init__(self, app: FastAPI):
        app.middleware("http")(self._middleware)
        self._app = app
        self._deps = Deps()

    async def _middleware(self, request: Request, call_next):
        dependency_context.set(self._deps)
        return await call_next(request)

    def register_with(self, reg: Callable[[Deps], Optional[Deps]]) -> Dependencies:
        reg(self._deps)
        return self

    def include_router(self, router: Union[APIRouter, DependencyRouter],
                       *,
                       prefix: str = "",
                       tags: Optional[List[Union[str, Enum]]] = None,
                       dependencies: Optional[Sequence[params.Depends]] = None,
                       default_response_class: Type[Response] = Default(
                           JSONResponse),
                       responses: Optional[Dict[Union[int,
                                                      str], Dict[str, Any]]] = None,
                       callbacks: Optional[List[BaseRoute]] = None,
                       deprecated: Optional[bool] = None,
                       include_in_schema: bool = True,
                       generate_unique_id_function: Callable[[APIRoute], str] = Default(
            utils.generate_unique_id
    ),) -> None:
        if isinstance(router, DependencyRouter):
            router.apply_dependencies(self._deps)
        self._app.include_router(
            router,
            prefix=prefix,
            tags=tags,
            dependencies=dependencies,
            default_response_class=default_response_class,
            responses=responses,
            callbacks=callbacks,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
            generate_unique_id_function=generate_unique_id_function,
        )


T = TypeVar("T")


def D(dependency: Type[T], *, use_cache: bool = True) -> T:
    def _resolver():
        val = dependency_context.get()
        if not isinstance(val, Deps):
            raise Exception(
                "D injector resolver must be used within request context")
        return val.get(dependency)
    return Depends(dependency=_resolver, use_cache=use_cache)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aslabs', 'aslabs.dependencies.fastapi']

package_data = \
{'': ['*']}

install_requires = \
['Flask>=2.0.3,<3.0.0', 'aslabs-dependencies']

setup_kwargs = {
    'name': 'aslabs-dependencies-fastapi',
    'version': '0.0.1',
    'description': '',
    'long_description': '## ASLabs Dependencies: FastAPI\n\nFastAPI Adapter for the ASLabs Dependencies package.\n\n# How to use it:\n\nIn your `main.py`, add the `Dependencies` additional handler:\n\n```py\nfrom fastapi import FastAPI\n\nfrom deps import Dependencies\n\n\napp = FastAPI()\ndeps = Dependencies(app)    # Add line\n```\n\nIn your various routers, replace `APIRouter` with `DependencyRouter` and use it to register your dependencies:\n\n```py\nfrom deps import DependencyRouter, D\n\n\nrouter = DependencyRouter()\nrouter.register_dependencies(lambda deps: deps.add_direct(MyDependencyClass))      # Register your dependencies here\n```\n\nOnce registered, use the `D` parameter function to request dependencies. Notably, you must specify the type as the parameter for `D` (cannot be infered).\n\n```py\n@router.get("/")\ndef index(dep: MyDependencyClass = D(MyDependencyClass)):\n    return {"Hello": dep.get_world()}\n```\n\nOnce setup, back in your `main.py`, register the router on `deps`, not on `app`. The same parameters can be used as `app.include_router`\n\n```py\ndeps.include_router(router)\n```\n\nYou are now done.\n\n## Adding global dependencies\n\nYou can add dependencies directly to the `deps` object, by calling `register_with`:\n\n```py\ndeps.register_with(lambda deps: deps.add_direct(GlobalDependency))\n```',
    'author': 'Titusz Ban',
    'author_email': 'tituszban@antisociallabs.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
